import json
import os
import subprocess
from pathlib import Path
from typing import Dict

import pytest

import laspy
from laspy.vlrs.geotiff import GeographicTypeGeoKey, ProjectedCSTypeGeoKey
from laspy.vlrs.known import GeoKeyDirectoryVlr
from tests.test_common import autzen_geo_proj_las, autzen_las, test1_4_las

try:
    import pyproj
except ModuleNotFoundError:
    pyproj = None


@pytest.fixture()
def file_wkt():
    return laspy.read(test1_4_las)


@pytest.fixture()
def file_geotiff():
    return laspy.read(autzen_las)


@pytest.fixture()
def file_geotiff_geo_proj():
    return laspy.read(autzen_geo_proj_las)


@pytest.fixture()
def file_with_both_wkt_and_geotiff_vlrs():
    path = Path(__file__).parent / "data" / "file_with_both_wkt_and_geotiff_vlrs.las"
    return laspy.read(path)


def has_pyproj() -> bool:
    return pyproj is not None


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_parse_crs_wkt(file_wkt):
    assert "Bound CRS" in file_wkt.header.parse_crs().type_name


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_parse_crs_geotiff(file_geotiff, file_geotiff_geo_proj):
    assert "Projected CRS" in file_geotiff.header.parse_crs().type_name
    assert (
        file_geotiff.header.parse_crs().to_epsg()
        == file_geotiff_geo_proj.header.parse_crs().to_epsg()
        == 2994
    )

    geokeys1 = get_geokeys(file_geotiff.header)
    geokeys2 = get_geokeys(file_geotiff_geo_proj.header)

    assert (
        geokeys1[ProjectedCSTypeGeoKey.id] == geokeys2[ProjectedCSTypeGeoKey.id] == 2994
    )
    assert GeographicTypeGeoKey.id not in geokeys1
    assert geokeys2[GeographicTypeGeoKey.id] == 4152


def get_geokeys(header: laspy.LasHeader) -> Dict:
    geo_vlr = header.vlrs.get_by_id("LASF_Projection")
    for rec in geo_vlr:
        if isinstance(rec, GeoKeyDirectoryVlr):
            return {k.id: k.value_offset for k in rec.geo_keys}


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_add_crs_wkt():
    header = laspy.LasHeader(point_format=6, version="1.4")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs)

    lasf_proj = header.vlrs.get_by_id("LASF_Projection")

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], "string")
    assert lasf_proj[0].string == crs.to_wkt()
    assert header.global_encoding.wkt


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_add_crs_geotiff():
    """
    Test adding geotiff crs seems ok
    """
    header = laspy.LasHeader(point_format=3, version="1.2")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs)

    lasf_proj = header.vlrs.get_by_id("LASF_Projection")

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], "geo_keys")
    assert hasattr(lasf_proj[1], "strings")
    assert header.global_encoding.wkt == False

    las = laspy.LasData(header=header)
    las = laspy.lib.write_then_read_again(las)
    epsg = las.header.parse_crs().to_epsg()
    assert epsg == 6432, "epsg after read-write is not the same"


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_add_crs_compatibility():
    header = laspy.LasHeader(point_format=3, version="1.4")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs, keep_compatibility=False)

    lasf_proj = header.vlrs.get_by_id("LASF_Projection")

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], "string")
    assert lasf_proj[0].string == crs.to_wkt()
    assert header.global_encoding.wkt


def has_pdal():
    try:
        subprocess.run(["pdal", "--version"], capture_output=True, check=True)
    except:
        return False
    else:
        return True


def get_pdal_srs(filename):
    process = subprocess.run(
        ["pdal", "info", "--metadata", filename], capture_output=True, check=True
    )
    json_output = json.loads(process.stdout.decode("utf-8"))["metadata"]["srs"]
    return json_output["wkt"], json_output["proj4"]


def geotiff_crs_pdal_test(crs):
    header = laspy.LasHeader(point_format=3, version="1.2")
    header.add_crs(crs)

    tmp_filename = f"delete_me_pdal_test_cs_{crs.to_epsg()}.las"

    las = laspy.LasData(header=header)
    las.write(tmp_filename)

    try:
        pdal_wkt, _ = get_pdal_srs(tmp_filename)

    finally:
        os.remove(tmp_filename)

    # We don't use pdal's proj4 as its seems not complete enough
    # to be 100% equal, while the wkt seems ok
    assert pyproj.CRS.from_wkt(pdal_wkt) == crs


@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_derived_projected_cs():
    crs = pyproj.CRS.from_epsg(3945)
    geotiff_crs_pdal_test(crs)


@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_geographic_cs():
    crs = pyproj.CRS.from_epsg(4326)
    geotiff_crs_pdal_test(crs)


@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_projected_cs():
    crs = pyproj.CRS.from_epsg(6432)
    geotiff_crs_pdal_test(crs)


@pytest.mark.skipif(not has_pyproj(), reason="pyproj not installed")
def test_preference_option(file_with_both_wkt_and_geotiff_vlrs):
    header = file_with_both_wkt_and_geotiff_vlrs.header

    expected_wkt_crs = """PROJCS["NAD83_2011_Nebraska_ft",GEOGCS["GCS_NAD_1983_2011",DATUM["D_NAD83_NATIONAL_SPATIAL_REFERENCE_SYSTEM_2011",SPHEROID["GRS_1980",6378137,298.257222101]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295]],PROJECTION["Lambert_Conformal_Conic"],PARAMETER["scale_factor",1],PARAMETER["standard_parallel_1",40],PARAMETER["standard_parallel_2",43],PARAMETER["central_meridian",-100],PARAMETER["latitude_of_origin",39.83333333333334],PARAMETER["false_easting",1640416.666666667],PARAMETER["false_northing",0],UNIT["Foot_US",0.30480060960121924]]"""
    expected_geotiff_crs = "epsg:32104"

    crs = header.parse_crs()
    assert str(crs) == expected_wkt_crs

    crs = header.parse_crs(prefer_wkt=True)
    assert str(crs) == expected_wkt_crs

    crs = header.parse_crs(prefer_wkt=False)
    assert str(crs).lower() == expected_geotiff_crs



import numpy as np
import pytest

import laspy

from ..test_common import skip_if_no_laz_backend
from . import skip_if_cli_deps_are_not_installed

skip_if_cli_deps_are_not_installed()

pytestmark = skip_if_no_laz_backend

from typer.testing import CliRunner

runner = CliRunner()
from laspy.cli.core import app


def test_cli_compress(tmp_path):
    input_path = "tests/data/simple.las"
    output_path = tmp_path / "simple.laz"
    result = runner.invoke(
        app,
        [
            "compress",
            input_path,
            "--output-path",
            str(output_path),
        ],
    )
    assert result.exit_code == 0, "{}".format(result.stdout)

    with laspy.open(output_path) as f:
        assert f.header.are_points_compressed is True

    original_las = laspy.read(input_path)
    compressed_las = laspy.read(output_path)

    assert original_las.point_format == compressed_las.point_format
    assert original_las.vlrs == compressed_las.vlrs
    assert np.all(original_las.points == compressed_las.points)


def test_cli_compress_cannot_overwrite(tmp_path):
    input_path = "tests/data/simple.las"
    output_path = tmp_path / "simple.laz"
    result = runner.invoke(
        app,
        [
            "compress",
            input_path,
            "--output-path",
            input_path,
        ],
    )
    assert result.exit_code != 0, "{}".format(result.stdout)


def test_cli_compress_non_existing_file(tmp_path):
    input_path = "tests/data/i-do-not-exist.las"
    output_path = tmp_path / "i-do-not-exist.laz"
    result = runner.invoke(
        app,
        [
            "compress",
            input_path,
            "--output-path",
            str(output_path),
        ],
    )
    assert result.exit_code != 0, "{}".format(result.stdout)


def test_cli_decompress(tmp_path):
    input_path = "tests/data/simple.laz"
    output_path = tmp_path / "simple.las"
    result = runner.invoke(
        app,
        [
            "decompress",
            input_path,
            "--output-path",
            str(output_path),
        ],
    )
    assert result.exit_code == 0, "{}".format(result.stdout)

    with laspy.open(output_path) as f:
        assert f.header.are_points_compressed is False

    original_las = laspy.read(input_path)
    compressed_las = laspy.read(output_path)

    assert original_las.point_format == compressed_las.point_format
    assert original_las.vlrs == compressed_las.vlrs
    assert np.all(original_las.points == compressed_las.points)


def test_cli_decompress_cannot_overwrite(tmp_path):
    input_path = "tests/data/simple.laz"
    output_path = tmp_path / "simple.las"
    result = runner.invoke(
        app,
        [
            "decompress",
            input_path,
            "--output-path",
            input_path,
        ],
    )
    assert result.exit_code != 0, "{}".format(result.stdout)


def test_cli_decompress_non_existing_file(tmp_path):
    input_path = "tests/data/i-do-not-exist.laz"
    output_path = tmp_path / "i-do-not-exist.las"
    result = runner.invoke(
        app,
        [
            "decompress",
            input_path,
            "--output-path",
            str(output_path),
        ],
    )
    assert result.exit_code != 0, "{}".format(result.stdout)



from pathlib import Path

import pytest

import laspy

SIMPLE_LAS_FILE_PATH = Path(__file__).parent / "data" / "simple.las"
VEGETATION1_3_LAS_FILE_PATH = Path(__file__).parent / "data" / "vegetation_1_3.las"
TEST1_4_LAS_FILE_PATH = Path(__file__).parent / "data" / "test1_4.las"
EXTRA_BYTES_LAS_FILE_PATH = Path(__file__).parent / "data" / "extrabytes.las"
LAS_1_4_WITH_EVLRS_FILE_PATH = Path(__file__).parent / "data" / "1_4_w_evlr.las"

SIMPLE_LAZ_FILE_PATH = Path(__file__).parent / "data" / "simple.laz"
EXTRA_BYTES_LAZ_FILE_PATH = Path(__file__).parent / "data" / "extra.laz"
PLANE_LAZ_FILE_PATH = Path(__file__).parent / "data" / "plane.laz"
AUTZEN_FILE_PATH = Path(__file__).parent / "data" / "autzen.las"
AUTZEN_GEO_PROJ_FILE_PATH = Path(__file__).parent / "data" / "autzen_geo_proj.las"
LAZ_1_4_WITH_EVLRS_FILE_PATH = Path(__file__).parent / "data" / "1_4_w_evlr.laz"

UNREGISTERED_EXTRA_BYTES_LAS = (
    Path(__file__).parent / "data" / "unregistered_extra_bytes.las"
)

ALL_LAS_FILE_PATH = [
    SIMPLE_LAS_FILE_PATH,
    VEGETATION1_3_LAS_FILE_PATH,
    TEST1_4_LAS_FILE_PATH,
    EXTRA_BYTES_LAS_FILE_PATH,
]

ALL_LAZ_FILE_PATH = [
    SIMPLE_LAZ_FILE_PATH,
    EXTRA_BYTES_LAZ_FILE_PATH,
    PLANE_LAZ_FILE_PATH,
]

ALL_LAZ_BACKEND = laspy.LazBackend.detect_available()

SUPPORTED_SINGULAR_EXTRA_BYTES_TYPE = [
    "u1",
    "u2",
    "u4",
    "u8",
    "i1",
    "i2",
    "i4",
    "i8",
    "f4",
    "f8",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
    "float32",
    "float64",
]

SUPPORTED_ARRAY_2_EXTRA_BYTES_TYPE = [
    f"2{base_type}" for base_type in SUPPORTED_SINGULAR_EXTRA_BYTES_TYPE
]

SUPPORTED_ARRAY_3_EXTRA_BYTES_TYPE = [
    f"3{base_type}" for base_type in SUPPORTED_SINGULAR_EXTRA_BYTES_TYPE
]

SUPPORTED_EXTRA_BYTES_TYPE = (
    SUPPORTED_SINGULAR_EXTRA_BYTES_TYPE
    + SUPPORTED_ARRAY_2_EXTRA_BYTES_TYPE
    + SUPPORTED_ARRAY_3_EXTRA_BYTES_TYPE
)


class NonSeekableStream:
    """
    Fake non stream / file object which simulates a file object
    on which we cannot seek
    """

    def __init__(self, inner):
        self.inner = inner

    def read(self, n):
        return self.inner.read(n)

    def seekable(self):
        return False

    def close(self):
        pass


@pytest.fixture()
def simple_las_path():
    return SIMPLE_LAS_FILE_PATH


@pytest.fixture(params=SUPPORTED_EXTRA_BYTES_TYPE)
def extra_bytes_params(request):
    return laspy.ExtraBytesParams(
        name="just_a_name", type=request.param, description="laspy test ExtraBytes"
    )


@pytest.fixture(params=[EXTRA_BYTES_LAS_FILE_PATH, EXTRA_BYTES_LAZ_FILE_PATH], ids=repr)
def las_file_path_with_extra_bytes(request):
    if request.param.suffix == ".laz" and len(laspy.LazBackend.detect_available()) == 0:
        return pytest.skip("No Laz Backend")
    else:
        return request.param


@pytest.fixture(params=ALL_LAS_FILE_PATH, ids=repr)
def las_file_path(request):
    return request.param


@pytest.fixture(params=ALL_LAZ_FILE_PATH, ids=repr)
def laz_file_path(request):
    if len(laspy.LazBackend.detect_available()) == 0:
        return pytest.skip("No Laz Backend")
    return request.param


@pytest.fixture(params=ALL_LAS_FILE_PATH + ALL_LAZ_FILE_PATH, ids=repr)
def file_path(request):
    if request.param.suffix == ".laz" and len(laspy.LazBackend.detect_available()) == 0:
        return pytest.skip("No Laz Backend")
    return request.param


@pytest.fixture(
    params=ALL_LAZ_BACKEND
    if ALL_LAZ_BACKEND
    else [pytest.mark.skip("No Laz Backend installed")]
)
def laz_backend(request):
    return request.param


@pytest.fixture()
def mmapped_file_path(tmp_path):
    import shutil

    copied_file = shutil.copy(SIMPLE_LAS_FILE_PATH, tmp_path)
    yield copied_file



import io
import os

import numpy as np
import pytest

import laspy
from tests.test_common import simple_las, simple_laz, skip_if_no_laz_backend


@pytest.fixture(
    params=[simple_las, simple_laz]
    if laspy.LazBackend.detect_available()
    else [simple_las],
    scope="session",
)
def read_simple(request):
    return laspy.read(request.param)


@pytest.fixture()
def open_simple():
    return open(simple_las, mode="rb")


@pytest.fixture()
def read_uncompressed():
    return laspy.read(simple_laz)


@pytest.fixture()
def get_header():
    with laspy.open(simple_las) as fin:
        return fin.header


def test_header(get_header):
    header = get_header
    assert header.file_source_id == 0
    assert header.version.major == 1
    assert header.version.minor == 2
    assert header.system_identifier == ""
    assert header.generating_software == "TerraScan"
    assert header.creation_date is None
    assert header.offset_to_point_data == 227
    assert len(header.vlrs) == 0
    assert header.point_format.id == 3
    assert header.point_format.size == 34
    assert header.point_count == 1065
    assert tuple(header.number_of_points_by_return[:5]) == (925, 114, 21, 5, 0)
    assert header.x_scale == 0.01
    assert header.y_scale == 0.01
    assert header.z_scale == 0.01
    assert header.x_offset == 0
    assert header.y_offset == 0
    assert header.z_offset == 0
    assert header.x_max == pytest.approx(638982.55)
    assert header.x_min == pytest.approx(635619.85)
    assert header.y_max == pytest.approx(853535.43)
    assert header.y_min == pytest.approx(848899.70)
    assert header.z_max == pytest.approx(586.38)
    assert header.z_min == pytest.approx(406.59)


def test_no_vlr_for_simple(read_simple):
    f = read_simple
    assert f.vlrs == []


def test_every_byte_has_been_read(open_simple):
    fp = open_simple
    _ = laspy.read(fp, closefd=False)
    assert fp.tell() == os.path.getsize(simple_las)
    fp.close()


def test_unscaled_x(read_simple):
    f = read_simple
    assert f.X.max() == 63898255
    assert f.X.min() == 63561985


def test_unscaled_y(read_simple):
    f = read_simple
    assert f.Y.max() == 85353543
    assert f.Y.min() == 84889970


def test_unscaled_z(read_simple):
    f = read_simple
    assert f.Z.max() == 58638
    assert f.Z.min() == 40659


def test_intensity(read_simple):
    f = read_simple
    assert f.intensity.max() == 254
    assert f.intensity.min() == 0


def test_return_number(read_simple):
    f = read_simple
    assert f.return_number.max() == 4
    assert f.return_number.min() == 1


def test_number_of_returns(read_simple):
    f = read_simple
    assert f.number_of_returns.max() == 4
    assert f.number_of_returns.min() == 1


def test_edge_of_flight_line(read_simple):
    f = read_simple
    assert f.edge_of_flight_line.max() == 0
    assert f.edge_of_flight_line.min() == 0


def test_scan_direction_flag(read_simple):
    f = read_simple
    assert f.scan_direction_flag.max() == 1
    assert f.scan_direction_flag.min() == 0


def test_scan_angle_rank(read_simple):
    f = read_simple
    assert f.scan_angle_rank.max() == 18
    assert f.scan_angle_rank.min() == -19


def test_classification_max_min(read_simple):
    f = read_simple
    assert f.classification.max() == 2
    assert f.classification.min() == 1


def test_classification_count(read_simple):
    f = read_simple
    uniques, counts = np.unique(f.classification, return_counts=True)
    assert np.all(uniques == [1, 2])
    assert counts[0] == 789  # class code 1
    assert counts[1] == 276  # class code 2


def test_user_data(read_simple):
    f = read_simple
    assert f.user_data.max() == 149
    assert f.user_data.min() == 117


def test_point_source_id(read_simple):
    f = read_simple
    assert f.point_source_id.max() == 7334
    assert f.point_source_id.min() == 7326


def test_gps_time(read_simple):
    f = read_simple
    assert f.gps_time.max() == pytest.approx(249783.162158)
    assert f.gps_time.min() == pytest.approx(245370.417075)


def test_red(read_simple):
    f = read_simple
    assert f.red.max() == 249
    assert f.red.min() == 39


def test_green(read_simple):
    f = read_simple
    assert f.green.max() == 239
    assert f.green.min() == 57


def test_blue(read_simple):
    f = read_simple
    assert f.blue.max() == 249
    assert f.blue.min() == 56


@skip_if_no_laz_backend
def test_decompression_is_same_as_uncompressed():
    u_las = laspy.read(simple_las)
    c_las = laspy.read(simple_laz)

    u_point_buffer = bytes(u_las.points.memoryview())
    c_points_buffer = bytes(c_las.points.memoryview())

    assert u_point_buffer == c_points_buffer


def check_seeking_works(las_reader: laspy.LasReader):
    p1 = las_reader.read_points(10)

    idx = las_reader.seek(5)
    assert idx == 5

    p2 = las_reader.read_points(5)
    assert p1[5:] == p2

    idx = las_reader.seek(-5, io.SEEK_CUR)
    assert idx == 5
    p2 = las_reader.read_points(5)
    assert p1[5:] == p2

    rest = las_reader.read_points(-1)
    assert len(rest) == las_reader.header.point_count - 10

    idx = las_reader.seek(-150, io.SEEK_END)
    assert idx == las_reader.header.point_count - 150
    p2 = las_reader.read_points(-1)
    assert p2 == rest[-150:]


def test_seek_las(las_file_path):
    with laspy.open(las_file_path) as reader:
        check_seeking_works(reader)


@pytest.mark.parametrize("laz_backend", laspy.LazBackend.detect_available())
def test_seek_laz(laz_file_path, laz_backend):
    with laspy.open(laz_file_path, laz_backend=laz_backend) as reader:
        check_seeking_works(reader)



import numpy as np
import pytest

import laspy

from . import skip_if_cli_deps_are_not_installed

skip_if_cli_deps_are_not_installed()

from typer.testing import CliRunner

from laspy.cli.core import (
    BinaryFilteringExpression,
    Comparator,
    Condition,
    FilteringAction,
    FilteringExpression,
    FilteringExpressionKind,
    NegatedFilteringExpression,
    Token,
    app,
    parse_float_or_int,
    parse_list_of_numbers,
    tokenize,
)

runner = CliRunner()


def test_parse_float_or_number():
    value = parse_float_or_int("10")
    assert type(value) == int
    assert value == 10

    value = parse_float_or_int(" 17 ")
    assert type(value) == int
    assert value == 17

    value = parse_float_or_int("0.3")
    assert type(value) == float
    assert value == 0.3

    value = parse_float_or_int(" 0.889 ")
    assert type(value) == float
    assert value == 0.889

    with pytest.raises(ValueError):
        _ = parse_float_or_int("238a")


def test_parse_list_of_numbers():
    value = parse_list_of_numbers("[1, 2, 3]")
    assert value == [1, 2, 3]

    value = parse_list_of_numbers("(4, 5, 6)")
    assert value == [4, 5, 6]

    with pytest.raises(ValueError):
        _ = parse_list_of_numbers("238")

    with pytest.raises(ValueError):
        _ = parse_list_of_numbers("[238, 1")

    with pytest.raises(ValueError):
        _ = parse_list_of_numbers("38, 1]")


def test_tokenize():
    tokens = tokenize("not classification == 2")

    Kind = Token.Kind
    expected = [
        Token(Kind.Not, "not"),
        Token(Kind.LiteralStr, "classification"),
        Token(Kind.EqEq, "=="),
        Token(Kind.LiteralStr, "2"),
    ]
    assert tokens == expected


def test_parse_valid_filter_action():
    action = FilteringAction.parse_string("intensity >= 128")

    expected = FilteringAction("intensity", Comparator.GreaterOrEqual, "128")
    assert action == expected


def test_parse_valid_filter_action_in():
    action = FilteringAction.parse_string("intensity in [128, 129, 130]")

    expected = FilteringAction("intensity", Comparator.In, "[128, 129, 130]")
    assert action == expected


def test_parse_invalid_filter_action_invalid_operator():
    with pytest.raises(ValueError):
        _ = FilteringAction.parse_string("point_source_id <> 128")


def test_parse_invalid_filter_action_no_operator():
    """This one show that our simplistic parsing catches some error late"""
    with pytest.raises(ValueError):
        _ = FilteringAction.parse_string("point_source_id 128")


def test_parse_filtering_expression():
    parsed_expr = FilteringExpression.parse_string(
        "classification == 2 or classification == 3 and x > 10.0"
    )

    """
                            ┌────┐
                            │ or │
                            └▲─▲─┘
                             │ │
                             │ │
                             │ │
       ┌─────────────────────┤ │
       │ classification == 2 │ │
       └─────────────────────┘ └───────┬─────┐
                                       │ and │
                                       └─▲──▲┘
                                         │  │
                  ┌─────────────────────┬┘  └─┬──────────┐
                  │ classification == 3 │     │ x > 10.0 │
                  └─────────────────────┘     └──────────┘
    """

    expected_expr = FilteringExpression(
        kind=FilteringExpressionKind.Binary,
        data=BinaryFilteringExpression(
            condition=Condition.Or,
            lhs=FilteringExpression(
                kind=FilteringExpressionKind.Action,
                data=FilteringAction(
                    field_name="classification",
                    comparator=Comparator.Equality,
                    value="2",
                ),
            ),
            rhs=FilteringExpression(
                kind=FilteringExpressionKind.Binary,
                data=BinaryFilteringExpression(
                    condition=Condition.And,
                    lhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="classification",
                            comparator=Comparator.Equality,
                            value="3",
                        ),
                    ),
                    rhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="x",
                            comparator=Comparator.GreaterThan,
                            value="10.0",
                        ),
                    ),
                ),
            ),
        ),
    )

    assert parsed_expr == expected_expr


def test_parse_filtering_expression_not():
    parsed_expr = FilteringExpression.parse_string("not classification == 2")

    expected_expr = FilteringExpression(
        kind=FilteringExpressionKind.Negated,
        data=NegatedFilteringExpression(
            expr=FilteringExpression(
                kind=FilteringExpressionKind.Action,
                data=FilteringAction(
                    field_name="classification",
                    comparator=Comparator.Equality,
                    value="2",
                ),
            ),
        ),
    )

    assert parsed_expr == expected_expr


def test_parse_filtering_expression_2():
    parsed_expr = FilteringExpression.parse_string(
        "(classification == 2 or classification == 3) and x > 10.0"
    )

    """
                                    ┌─────┐
                                    │ And │
                                    └─▲─▲─┘
                                      │ │
                                      │ │
                         ┌────┬───────┘ │
                  ┌─────►│ or │         └────────────────┬──────────┐
                  │      └───▲┘                          │ x > 10.0 │
                  │          │                           └──────────┘
   ┌──────────────┴───────┐ ┌┴────────────────────┐
   │ classification == 2  │ │ classification == 3 │
   └──────────────────────┘ └─────────────────────┘
    """

    expected_expr = FilteringExpression(
        kind=FilteringExpressionKind.Binary,
        data=BinaryFilteringExpression(
            condition=Condition.And,
            lhs=FilteringExpression(
                kind=FilteringExpressionKind.Binary,
                data=BinaryFilteringExpression(
                    condition=Condition.Or,
                    lhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="classification",
                            comparator=Comparator.Equality,
                            value="2",
                        ),
                    ),
                    rhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="classification",
                            comparator=Comparator.Equality,
                            value="3",
                        ),
                    ),
                ),
            ),
            rhs=FilteringExpression(
                kind=FilteringExpressionKind.Action,
                data=FilteringAction(
                    field_name="x",
                    comparator=Comparator.GreaterThan,
                    value="10.0",
                ),
            ),
        ),
    )

    assert parsed_expr == expected_expr


def test_parse_filtering_expression_2_pdal_style():
    parsed_expr = FilteringExpression.parse_string(
        "(classification == 2 || classification == 3) && x > 10.0"
    )

    expected_expr = FilteringExpression(
        kind=FilteringExpressionKind.Binary,
        data=BinaryFilteringExpression(
            condition=Condition.And,
            lhs=FilteringExpression(
                kind=FilteringExpressionKind.Binary,
                data=BinaryFilteringExpression(
                    condition=Condition.Or,
                    lhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="classification",
                            comparator=Comparator.Equality,
                            value="2",
                        ),
                    ),
                    rhs=FilteringExpression(
                        kind=FilteringExpressionKind.Action,
                        data=FilteringAction(
                            field_name="classification",
                            comparator=Comparator.Equality,
                            value="3",
                        ),
                    ),
                ),
            ),
            rhs=FilteringExpression(
                kind=FilteringExpressionKind.Action,
                data=FilteringAction(
                    field_name="x",
                    comparator=Comparator.GreaterThan,
                    value="10.0",
                ),
            ),
        ),
    )

    assert parsed_expr == expected_expr


def test_apply_filter_expression():
    parsed_expr = FilteringExpression.parse_string(
        "classification == 2 or classification == 3 and x > 10.0"
    )

    points = laspy.ScaleAwarePointRecord.zeros(
        3,
        point_format=laspy.PointFormat(3),
        scales=np.array([1.0, 1.0, 1.0]),
        offsets=np.array([1.0, 1.0, 1.0]),
    )
    points["classification"] = [1, 2, 3]
    points["x"] = [1.0, 5.0, 11.0]

    result = parsed_expr.apply(points)

    expected = np.array([False, True, True])
    assert np.all(result == expected)


def test_apply_filter_expression_2():
    parsed_expr = FilteringExpression.parse_string(
        "(classification == 2 or classification == 3) and x > 10.0"
    )

    points = laspy.ScaleAwarePointRecord.zeros(
        3,
        point_format=laspy.PointFormat(3),
        scales=np.array([1.0, 1.0, 1.0]),
        offsets=np.array([1.0, 1.0, 1.0]),
    )
    points["classification"] = [1, 2, 3]
    points["x"] = [1.0, 5.0, 11.0]

    result = parsed_expr.apply(points)

    expected = np.array([False, False, True])
    assert np.all(result == expected)


def test_cli_filtering_classification_eq_2(tmp_path):
    path = tmp_path / "simple_cls_2.las"
    result = runner.invoke(
        app, ["filter", "tests/data/simple.las", str(path), "classification == 2"]
    )
    assert result.exit_code == 0

    las = laspy.read(path)
    assert np.all(las.classification == 2)


def test_cli_filtering_classification_ne_2(tmp_path):
    path = tmp_path / "simple_cls_2.las"
    result = runner.invoke(
        app, ["filter", "tests/data/simple.las", str(path), "classification != 2"]
    )
    assert result.exit_code == 0

    las = laspy.read(path)
    assert np.all(las.classification != 2)


def test_cli_filtering_classification_in(tmp_path):
    path = tmp_path / "simple_cls_2.las"
    result = runner.invoke(
        app,
        [
            "filter",
            "tests/data/simple.las",
            str(path),
            "classification in [2, 3, 4]",
        ],
    )
    assert result.exit_code == 0

    las = laspy.read(path)
    assert np.alltrue(np.isin(las.classification, [2, 3, 4]))


def test_cli_filtering_classification_intensity(tmp_path):
    path = tmp_path / "simple_cls_2.las"
    result = runner.invoke(
        app,
        [
            "filter",
            "tests/data/simple.las",
            str(path),
            "(intensity >= 128 and intensity <= 256) and classification == 2",
        ],
    )
    assert result.exit_code == 0

    las = laspy.read(path)
    assert np.all(las.classification == 2)
    assert np.all(las.intensity >= 128)
    assert np.all(las.intensity <= 256)


def test_cli_filtering(tmp_path):
    path = tmp_path / "simple_cls_2.las"
    result = runner.invoke(
        app,
        [
            "filter",
            "tests/data/simple.las",
            str(path),
            "intensity <= 42 and not classification == 2",
        ],
    )
    assert result.exit_code == 0

    las = laspy.read(path)
    assert np.all(las.intensity <= 42)
    assert np.all(las.classification != 2)



import io
import subprocess
import sys
from pathlib import Path

import numpy as np
import pytest

import laspy

SIMPLE_COPC_FILE = Path(__file__).parent / "data" / "simple.copc.laz"

try:
    import requests
except ModuleNotFoundError:
    requests = None

try:
    import RangeHTTPServer
except ModuleNotFoundError:
    RangeHTTPServer = None


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(
            laspy.LazBackend.Laszip,
            marks=pytest.mark.skipif("not laspy.LazBackend.Laszip.is_available()"),
        ),
        pytest.param(
            laspy.LazBackend.Lazrs,
            marks=pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()"),
        ),
        pytest.param(
            laspy.LazBackend.LazrsParallel,
            marks=pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()"),
        ),
    ],
)
def test_reading_copc_file_normal_laz_file(backend):
    # COPC files are LAZ files with data arranged
    # in a special way that is still compatible with
    # standard file.
    # So we must be able to read a copc file as if it was
    # a classical LAZ

    las = laspy.read(SIMPLE_COPC_FILE, laz_backend=backend)
    assert las.header.version == "1.4"
    assert las.header.point_format == laspy.PointFormat(7)
    assert len(las) == 1065


@pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file():
    with laspy.CopcReader.open(SIMPLE_COPC_FILE) as copc_reader:
        assert copc_reader.header.version == "1.4"
        assert copc_reader.header.point_format == laspy.PointFormat(7)
        points = copc_reader.query(resolution=50)
        assert len(points) == 24


@pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file_with_page():
    path = SIMPLE_COPC_FILE.with_name("simple_with_page.copc.laz")
    point_count = laspy.read(path).header.point_count
    with laspy.CopcReader.open(path) as copc_reader:
        points = copc_reader.query()
        assert point_count == 1065 == len(points)


@pytest.mark.skipif("laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file_proper_error_if_no_lazrs():
    with pytest.raises(laspy.errors.LazError):
        with laspy.CopcReader.open(SIMPLE_COPC_FILE) as _:
            pass


@pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file_object():
    with open(SIMPLE_COPC_FILE, "rb") as fh:
        with laspy.CopcReader.open(fh) as copc_reader:
            assert copc_reader.header.version == "1.4"
            assert copc_reader.header.point_format == laspy.PointFormat(7)
            points = copc_reader.query(resolution=50)
            assert len(points) == 24


@pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file_object_with_page():
    path = SIMPLE_COPC_FILE.with_name("simple_with_page.copc.laz")
    point_count = laspy.read(path).header.point_count
    with open(path, "rb") as fh:
        with laspy.CopcReader.open(fh) as copc_reader:
            points = copc_reader.query()
            assert point_count == 1065 == len(points)


@pytest.mark.skipif("laspy.LazBackend.Lazrs.is_available()")
def test_querying_copc_local_file_object_proper_error_if_no_lazrs():
    with pytest.raises(laspy.errors.LazError):
        with open(SIMPLE_COPC_FILE, "rb") as fh:
            with laspy.CopcReader.open(fh) as _:
                pass


@pytest.mark.skipif(
    not (
        laspy.LazBackend.Lazrs.is_available()
        and requests is not None
        and RangeHTTPServer is not None
    ),
    reason="neither lazrs, nor requests, nor RangeHTTPServer are installed",
)
def test_copc_over_http():
    server_proc = subprocess.Popen(
        [sys.executable, "-m", "RangeHTTPServer"], cwd=str(Path(__file__).parent)
    )

    with laspy.CopcReader.open(
        "http://localhost:8000/data/simple.copc.laz"
    ) as copc_reader:
        assert copc_reader.header.version == "1.4"
        assert copc_reader.header.point_format == laspy.PointFormat(7)
        points = copc_reader.query(resolution=50)
        assert len(points) == 24

    server_proc.terminate()


@pytest.mark.parametrize(
    "exc_type,backend",
    [
        pytest.param(
            laspy.LaspyException,
            laspy.LazBackend.Laszip,
            marks=pytest.mark.skipif("not laspy.LazBackend.Laszip.is_available()"),
        ),
        pytest.param(
            NotImplementedError,
            laspy.LazBackend.Lazrs,
            marks=pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()"),
        ),
        pytest.param(
            NotImplementedError,
            laspy.LazBackend.LazrsParallel,
            marks=pytest.mark.skipif("not laspy.LazBackend.Lazrs.is_available()"),
        ),
    ],
)
def test_writing_copc_file_fails(exc_type, backend):
    las = laspy.read(SIMPLE_COPC_FILE, laz_backend=backend)

    with pytest.raises(exc_type):
        with io.BytesIO() as output:
            las.write(output, laz_backend=backend)


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(
            laspy.LazBackend.Lazrs,
            marks=pytest.mark.skipif(
                not laspy.LazBackend.Lazrs.is_available(), reason="lazrs not installed"
            ),
        ),
        pytest.param(
            laspy.LazBackend.LazrsParallel,
            marks=pytest.mark.skipif(
                not laspy.LazBackend.Lazrs.is_available(), reason="lazrs not installed"
            ),
        ),
    ],
)
def test_copc_selective_decompression(backend):
    # We only decompress X,Y, return number, number of returns, classification
    selection = laspy.DecompressionSelection.base().decompress_classification()

    with laspy.CopcReader.open(SIMPLE_COPC_FILE) as copc_reader:
        fully_decompressed_points = copc_reader.query()

    with laspy.CopcReader.open(
        SIMPLE_COPC_FILE, decompression_selection=selection
    ) as copc_reader:
        partially_decompressed_points = copc_reader.query()

    print(
        np.sum(
            fully_decompressed_points.classification
            == partially_decompressed_points.classification
        ),
        len(fully_decompressed_points),
    )

    assert np.all(fully_decompressed_points.X == partially_decompressed_points.X)
    assert np.all(fully_decompressed_points.Y == partially_decompressed_points.Y)
    assert np.all(
        fully_decompressed_points.return_number
        == partially_decompressed_points.return_number
    )
    assert np.all(
        fully_decompressed_points.number_of_returns
        == partially_decompressed_points.number_of_returns
    )
    # assert np.all(
    #     fully_decompressed_points.classification == partially_decompressed_points.classification
    # )

    # Since COPC uses variable chunk size its easier to test that
    # values are all not eq between fully and partially decompressed
    assert np.any(
        partially_decompressed_points.point_source_id
        != fully_decompressed_points.point_source_id
    )
    assert np.any(partially_decompressed_points.Z != fully_decompressed_points.Z)



import os
from pathlib import Path

import numpy as np
import pytest

import laspy
from laspy import LazBackend
from tests.test_common import test1_4_las


@pytest.fixture()
def file():
    return laspy.read(test1_4_las)


def test_unscaled_x(file):
    assert file.X.max() == 1751224820
    assert file.X.min() == 1320803567


def test_unscaled_y(file):
    assert file.Y.max() == -860121188
    assert file.Y.min() == -864646690


def test_unscaled_z(file):
    assert file.Z.max() == -1745638014
    assert file.Z.min() == -1751937981


def test_intensity(file):
    assert file.intensity.max() == 68
    assert file.intensity.min() == 2


def test_return_number(file):
    assert file.return_number.max() == 4
    assert file.return_number.min() == 1


def test_number_of_returns(file):
    assert file.number_of_returns.max() == 4
    assert file.number_of_returns.min() == 1


def test_edge_of_flight_line(file):
    assert file.edge_of_flight_line.max() == 1
    assert file.edge_of_flight_line.min() == 0


def scan_direction_flag(file):
    assert file.scan_direction_flag.max() == 1
    assert file.scan_direction_flag.min() == 0


def test_classification(file):
    assert file.classification.max() == 2
    assert file.classification.min() == 2


def test_scan_angle_rank(file):
    assert file.scan_angle.max() == 3173
    assert file.scan_angle.min() == 1837


def test_user_data(file):
    assert file.user_data.max() == 0
    assert file.user_data.min() == 0


def test_point_source_id(file):
    assert file.point_source_id.max() == 202
    assert file.point_source_id.min() == 202


def test_gps_time(file):
    assert file.gps_time.max() == pytest.approx(83177420.601045)
    assert file.gps_time.min() == pytest.approx(83177420.534005)


def test_scanner_channel(file):
    assert file.scanner_channel.max() == 0
    assert file.scanner_channel.min() == 0


def test_we_read_evlrs_when_simply_opening():
    file_with_evlrs = os.path.dirname(__file__) + "/data/1_4_w_evlr.las"
    expected_evlrs = laspy.VLR(
        user_id="pylastest",
        record_id=42,
        description="just a test evlr",
        record_data=b"Test 1 2 ... 1 2",
    )
    with laspy.open(file_with_evlrs) as reader:
        assert reader.evlrs == [expected_evlrs]


def test_we_dont_read_evlrs_when_opening_if_user_does_not_want():
    file_with_evlrs = os.path.dirname(__file__) + "/data/1_4_w_evlr.las"
    with laspy.open(file_with_evlrs, read_evlrs=False) as reader:
        assert reader.evlrs is None


def test_reader_read_reads_evlrs_even_if_skipped_at_opening():
    file_with_evlrs = os.path.dirname(__file__) + "/data/1_4_w_evlr.las"
    expected_evlrs = [
        laspy.VLR(
            user_id="pylastest",
            record_id=42,
            description="just a test evlr",
            record_data=b"Test 1 2 ... 1 2",
        )
    ]
    with laspy.open(file_with_evlrs, read_evlrs=False) as reader:
        assert reader.evlrs is None
        las = reader.read()
        assert las.evlrs == expected_evlrs
        assert reader.evlrs == expected_evlrs


@pytest.mark.parametrize(
    "file,backend",
    [
        ("/data/1_4_w_evlr.las", None),
        pytest.param(
            "/data/1_4_w_evlr.laz",
            LazBackend.Laszip,
            marks=pytest.mark.skipif(
                not LazBackend.Laszip.is_available(), reason="laszip is not installed"
            ),
        ),
        pytest.param(
            "/data/1_4_w_evlr.laz",
            LazBackend.Lazrs,
            marks=pytest.mark.skipif(
                not LazBackend.Lazrs.is_available(), reason="lazrs is not installed"
            ),
        ),
        pytest.param(
            "/data/1_4_w_evlr.laz",
            LazBackend.LazrsParallel,
            marks=pytest.mark.skipif(
                not LazBackend.LazrsParallel.is_available(),
                reason="lazrs is not installed",
            ),
        ),
    ],
)
def test_manually_reading_evlrs(file, backend):
    # The goal is to test we can read evlrs
    # in between reading points, as reading evlrs
    # will require to seek to them, so this is to make sure
    # we correctly reseek to the previous position
    # and that we can continue to read points

    file_with_evlrs = os.path.dirname(__file__) + file
    expected_evlrs = [
        laspy.VLR(
            user_id="pylastest",
            record_id=42,
            description="just a test evlr",
            record_data=b"Test 1 2 ... 1 2",
        )
    ]

    expected_points = laspy.read(file_with_evlrs, laz_backend=backend).points

    with laspy.open(file_with_evlrs, read_evlrs=False, laz_backend=backend) as reader:
        points = laspy.ScaleAwarePointRecord.empty(header=reader.header)

        assert reader.evlrs is None

        chunk_size = 100

        iter = reader.chunk_iterator(chunk_size)
        chunk1 = next(iter)
        assert np.alltrue(chunk1 == expected_points[:chunk_size])

        reader.read_evlrs()
        assert reader.evlrs == expected_evlrs

        for i, chunk in enumerate(iter):
            assert np.alltrue(
                chunk == expected_points[(i + 1) * chunk_size : (i + 2) * chunk_size]
            )


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(
            laspy.LazBackend.Laszip,
            marks=pytest.mark.skipif(
                not laspy.LazBackend.Laszip.is_available(),
                reason="Laszip not installed",
            ),
        ),
        pytest.param(
            laspy.LazBackend.Lazrs,
            marks=pytest.mark.skipif(
                not laspy.LazBackend.Lazrs.is_available(), reason="lazrs not installed"
            ),
        ),
        pytest.param(
            laspy.LazBackend.LazrsParallel,
            marks=pytest.mark.skipif(
                not laspy.LazBackend.Lazrs.is_available(), reason="lazrs not installed"
            ),
        ),
    ],
)
def test_selective_decompression(backend):
    # We only decompress X,Y, return number, number of returns, intensity
    selection = laspy.DecompressionSelection.base().decompress_intensity()

    simple1_4 = Path(__file__).parent / "data" / "1_4_w_evlr.laz"

    fully_decompressed_laz = laspy.read(simple1_4, laz_backend=backend)
    partially_decompressed_laz = laspy.read(
        simple1_4, laz_backend=backend, decompression_selection=selection
    )

    assert fully_decompressed_laz.point_format.id >= 6

    assert np.all(fully_decompressed_laz.X == partially_decompressed_laz.X)
    assert np.all(fully_decompressed_laz.Y == partially_decompressed_laz.Y)
    assert np.all(
        fully_decompressed_laz.return_number == partially_decompressed_laz.return_number
    )
    assert np.all(
        fully_decompressed_laz.number_of_returns
        == partially_decompressed_laz.number_of_returns
    )
    assert np.all(
        fully_decompressed_laz.intensity == partially_decompressed_laz.intensity
    )

    # This is the chunk size of the chunks inside the .laz
    chunk_size = 50_000
    # In selective decompression, the bytes which are not decompressed are set to the value
    # of the first point of the LAZ chunk
    for i in range((len(fully_decompressed_laz) // chunk_size) + 1):
        start, end = i, (i + 1) * chunk_size
        assert np.all(
            # 12
            fully_decompressed_laz.classification[start]
            == partially_decompressed_laz.classification[start:end]
        )
        assert np.all(
            fully_decompressed_laz.Z[start] == partially_decompressed_laz.Z[start:end]
        )



import pytest

from . import skip_if_cli_deps_are_not_installed

skip_if_cli_deps_are_not_installed()

from typer.testing import CliRunner

from laspy.cli.core import app

runner = CliRunner()

EXPECTED_HEADER_INFO_SIMPLE_LAS = (
    "                                 Header                                  \n"
    " Version                     1.2                                         \n"
    " Point Format Id             3                                           \n"
    " Point Format Size           34                                          \n"
    " Extra Bytes                 0                                           \n"
    " Point Count                 1065                                        \n"
    " Compressed                  False                                       \n"
    " System Identifier           ''                                          \n"
    " Generating Software         'TerraScan'                                 \n"
    " Number Of VLRs              0                                           \n"
    " UUID                        '00000000-0000-0000-0000-000000000000'      \n"
    " File Source Id              0                                           \n"
    " Creation Date               None                                        \n"
    " Scales                      [0.01 0.01 0.01]                            \n"
    " Offsets                     [-0. -0. -0.]                               \n"
    " Mins                        [6.3561985e+05 8.4889970e+05 4.0659000e+02] \n"
    " Maxs                        [6.3898255e+05 8.5353543e+05 5.8638000e+02] \n"
    " Number Of Points By Return  [925 114  21   5   0]                       \n"
)

EXPECTED_VLR_INFO_SIMPLE_LAS = ""

EXPECTED_POINTS_INFO_SIMPLE_LAS = (
    "\n"
    "                    Stats                     \n"
    " Dimension Name       Min  Max                \n"
    " X                    0    63898255           \n"
    " Y                    0    85353543           \n"
    " Z                    0    58638              \n"
    " intensity            0    254                \n"
    " return_number        0    4                  \n"
    " number_of_returns    0    4                  \n"
    " scan_direction_flag  0    1                  \n"
    " edge_of_flight_line  0    0                  \n"
    " classification       0    2                  \n"
    " synthetic            0    0                  \n"
    " key_point            0    0                  \n"
    " withheld             0    0                  \n"
    " scan_angle_rank      -19  18                 \n"
    " user_data            0    149                \n"
    " point_source_id      0    7334               \n"
    " gps_time             0.0  249783.16215837188 \n"
    " red                  0    249                \n"
    " green                0    239                \n"
    " blue                 0    249                \n"
)

EXPECTED_INFO_SIMPLE_LAS = (
    f"{EXPECTED_HEADER_INFO_SIMPLE_LAS}"
    "--------------------------------------------------\n"
    f"{EXPECTED_VLR_INFO_SIMPLE_LAS}"
    "--------------------------------------------------\n"
    f"{EXPECTED_POINTS_INFO_SIMPLE_LAS}"
)

EXPECTED_VLR_INFO_AUTZEN_LAS = (
    "                            VLRs                            \n"
    " User ID          Record ID  Description                    \n"
    " liblas           2112       OGR variant of OpenGIS WKT SRS \n"
    " LASF_Projection  34735      GeoTIFF GeoKeyDirectoryTag     \n"
    " LASF_Projection  34737      GeoTIFF GeoAsciiParamsTag      \n"
    " liblas           2112       OGR variant of OpenGIS WKT SRS \n"
)


def test_header_info():
    result = runner.invoke(app, ["info", "--header", "tests/data/simple.las"])
    assert result.exit_code == 0

    output = result.stdout
    assert output == EXPECTED_HEADER_INFO_SIMPLE_LAS


def test_vlr_info():
    result = runner.invoke(app, ["info", "--vlrs", "tests/data/simple.las"])
    assert result.exit_code == 0
    assert result.stdout == EXPECTED_VLR_INFO_SIMPLE_LAS

    result = runner.invoke(app, ["info", "--vlrs", "tests/data/autzen.las"])
    assert result.exit_code == 0

    assert result.stdout == EXPECTED_VLR_INFO_AUTZEN_LAS


def test_point_info():
    result = runner.invoke(app, ["info", "--points", "tests/data/simple.las"])
    assert result.exit_code == 0

    assert result.stdout == EXPECTED_POINTS_INFO_SIMPLE_LAS


def test_complete_info():
    result = runner.invoke(app, ["info", "tests/data/simple.las"])
    assert result.exit_code == 0

    assert result.stdout == EXPECTED_INFO_SIMPLE_LAS


def test_info_non_existant_file():
    result = runner.invoke(
        app, ["info", "--header", "tests/data/this_does_not_exist.las"]
    )
    assert result.exit_code != 0
    assert (
        result.output
        == """Error:
[Errno 2] No such file or directory: 'tests/data/this_does_not_exist.las'
"""
        or result.output  # Windows paths...
        == """Error:
[Errno 2] No such file or directory: 'tests\\\\data\\\\this_does_not_exist.las'
"""
    )



import numpy as np

import laspy


def test_mmap(mmapped_file_path):
    with laspy.mmap(mmapped_file_path) as las:
        las.classification[:] = 25
        assert np.all(las.classification == 25)

    las = laspy.read(mmapped_file_path)
    assert np.all(las.classification == 25)



"""
Tests related to the 'chunked' reading and writing
"""
import io
import math

import numpy as np
import pytest

import laspy


def check_chunked_reading_gives_expected_points(
    groundtruth_las: laspy.LasData, reader: laspy.LasReader, iter_size: int
):
    """Checks that the points read by the reader are the same as groundtruth points."""
    assert groundtruth_las.point_format == reader.header.point_format
    for i, points in enumerate(reader.chunk_iterator(iter_size)):
        expected_points = groundtruth_las.points[i * iter_size : (i + 1) * iter_size]
        for dim_name in points.array.dtype.names:
            assert np.allclose(
                expected_points[dim_name], points[dim_name]
            ), f"{dim_name} not equal"


def test_chunked_las_reading_gives_expected_points(las_file_path):
    """
    Test chunked LAS reading
    """
    with laspy.open(las_file_path) as las_reader:
        with laspy.open(las_file_path) as reader:
            las = las_reader.read()
            check_chunked_reading_gives_expected_points(las, reader, iter_size=50)


def test_chunked_laz_reading_gives_expected_points(laz_file_path, laz_backend):
    """
    Test LAZ reading in chunked mode with different backends
    """
    with laspy.open(laz_file_path) as las_reader:
        with laspy.open(laz_file_path, laz_backend=laz_backend) as laz_reader:
            expected_las = las_reader.read()
            check_chunked_reading_gives_expected_points(
                expected_las, laz_reader, iter_size=50
            )


@pytest.mark.parametrize("backend", laspy.LazBackend.detect_available() + (None,))
def test_chunked_writing_gives_expected_points(file_path, backend):
    """
    Write in chunked mode then test that the points are correct
    """
    original_las = laspy.read(file_path)
    iter_size = 51

    do_compress = True if backend is not None else False

    with io.BytesIO() as tmp_output:
        with laspy.open(
            tmp_output,
            mode="w",
            closefd=False,
            header=original_las.header,
            do_compress=do_compress,
            laz_backend=backend,
        ) as las:
            for i in range(int(math.ceil(len(original_las.points) / iter_size))):
                original_points = original_las.points[
                    i * iter_size : (i + 1) * iter_size
                ]
                las.write_points(original_points)

        tmp_output.seek(0)
        with laspy.open(tmp_output, closefd=False) as reader:
            check_chunked_reading_gives_expected_points(original_las, reader, iter_size)


@pytest.mark.parametrize("backend", laspy.LazBackend.detect_available() + (None,))
def test_chunked_dimension_modification(file_path, backend):
    """
    Test that when using chunked mode, if we modify the values in the chunks
    before writing them the values are actually modified.
    """

    mem_dest = io.BytesIO()

    # Read of file in chunked mode modify some values before writing them
    # to a new file
    with laspy.open(file_path, laz_backend=backend) as reader:
        with laspy.open(
            mem_dest, mode="w", laz_backend=backend, header=reader.header, closefd=False
        ) as dest:
            for points in reader.chunk_iterator(50):
                # try a sub field
                points.classification += 1
                # try a normal field
                points.user_data += 2
                # try a scaled field
                points.x += 10.0

                dest.write_points(points)

    mem_dest.seek(0)

    # Prepare out ground truth
    ground_truth_las = laspy.read(file_path)
    ground_truth_las.classification += 1
    ground_truth_las.user_data += 2
    ground_truth_las.x += 10.0

    reader = laspy.open(mem_dest, laz_backend=backend)
    check_chunked_reading_gives_expected_points(ground_truth_las, reader, 42)



import io
from pathlib import Path

import pytest

import laspy
from tests import test_common


def test_adding_classification_lookup():
    simple = laspy.read(test_common.simple_las)
    classification_lookup = laspy.vlrs.known.ClassificationLookupVlr()

    assert len(classification_lookup.lookups) == 0
    classification_lookup[20] = "computer"
    assert len(classification_lookup.lookups) == 1
    classification_lookup[17] = "car"

    simple.vlrs.append(classification_lookup)

    simple = test_common.write_then_read_again(simple)
    classification_lookups = simple.vlrs.get("ClassificationLookupVlr")[0]

    assert classification_lookups[20] == "computer"
    assert classification_lookups[17] == "car"


def test_lookup_out_of_range():
    classification_lookup = laspy.vlrs.known.ClassificationLookupVlr()
    with pytest.raises(ValueError):
        classification_lookup[541] = "LiquidWater"

    with pytest.raises(ValueError):
        classification_lookup[-42] = "SolidWater"


def test_adding_extra_bytes_vlr_by_hand():
    """
    Test that if someone adds an ExtraBytesVlr by himself
    without having matching extra bytes in the point record, the
    ExtraByteVlr is removed before writing
    """

    simple = laspy.read(test_common.simple_las)
    ebvlr = laspy.vlrs.known.ExtraBytesVlr()
    ebs = laspy.vlrs.known.ExtraBytesStruct(data_type=3, name="Fake".encode())
    ebvlr.extra_bytes_structs.append(ebs)
    simple.vlrs.append(ebvlr)
    assert len(simple.vlrs.get("ExtraBytesVlr")) == 1

    las = laspy.lib.write_then_read_again(simple)
    assert simple.points.point_size == las.points.point_size
    assert len(las.vlrs.get("ExtraBytesVlr")) == 0


def test_geokey_parsing_does_not_require_optional_params():
    las = laspy.read(str(Path(__file__).parent / "data/simple1_3.las"))
    geo_keys = laspy.vlrs.geotiff.parse_geo_tiff_keys_from_vlrs(las.vlrs)
    assert len(geo_keys) == 6


def test_cannot_write_vlrs_with_more_than_uint16_max_bytes():
    las = laspy.read(test_common.simple_las)
    big_junk_vlr = laspy.VLR(
        user_id="LASPY_ID",
        record_id=0,
        description="A VLR full of junk data",
        record_data=b"1" * (65_535 + 1),
    )
    las.vlrs.append(big_junk_vlr)

    with pytest.raises(ValueError):
        with io.BytesIO() as output:
            las.write(output)



import numpy as np
import pytest

import laspy
from laspy.point.dims import ScaledArrayView, SubFieldView


def test_sub_field_view_behaves_like_array():
    """This function is used to test if the SubFieldView class
    works & has an API that is similar to np.ndarray
    """
    array = np.zeros(10, np.uint8)

    field = SubFieldView(array, 0b0000_0010)

    assert len(field) == 10
    assert np.all(field == 0)
    assert np.all(field[:] == 0)

    assert field.max() == 0
    assert np.max(field) == 0
    assert field.min() == 0
    assert np.min(field) == 0
    assert np.min(field) == field.min()
    assert np.max(field) == field.max()

    field[:] = 1
    assert np.all(field == 1)
    assert np.all(field[:] == 1)

    assert field.max() == 1
    assert np.max(field) == 1
    assert field.min() == 1
    assert np.min(field) == 1

    assert np.all(field > 0)
    assert not np.all(field < 0)
    assert np.all(field >= 1)
    assert np.all(field <= 1)

    # check that the real array is properly modified
    assert np.all(array == 2)

    with pytest.raises(OverflowError):
        field[4] = 2

    assert np.mean(field) == 1


def test_array_view_int_index_return_singular_elements():
    a = np.array([1, 2, 3, 4], np.int32)
    s = SubFieldView(a, 0x00_00_00_FF)

    for i in range(len(s)):
        assert type(s[i]) in (np.int32, np.int64)
        assert a[i] == s[i]

    s = ScaledArrayView(a, scale=2.0, offset=0.0)
    for i in range(len(s)):
        assert type(s[i]) == np.float64
        assert (a[i] * 2.0) == s[i]


def test_scaled_array_view_ellipsis_indexing(simple_las_path):
    las = laspy.read(simple_las_path)

    las.add_extra_dim(
        laspy.ExtraBytesParams(
            name="test_dim",
            type="3int32",
            scales=np.array([1.0, 2.0, 3.0], np.float64),
            offsets=np.array([10.0, 20.0, 30.0], np.float64),
        )
    )

    # Query all the points in the 2nd dimension
    assert np.all(las.test_dim[..., 2] == 30.0)

    # Query the 10 nth point, we expect all its dimensions
    assert np.all(las.test_dim[10, ...] == [10.0, 20.0, 30.0])


def test_scaled_array_view_indexing_with_array_or_list(simple_las_path):
    las = laspy.read(simple_las_path)

    las.add_extra_dim(
        laspy.ExtraBytesParams(
            name="test_dim",
            type="3int32",
            scales=np.array([1.0, 2.0, 3.0], np.float64),
            offsets=np.array([10.0, 20.0, 30.0], np.float64),
        )
    )

    d = las.test_dim[[0, 1, 10, 12]]
    assert d.ndim == 2
    assert d.shape == (4, 3)
    assert np.all(d[..., 0] == 10.0)
    assert np.all(d[..., 1] == 20.0)
    assert np.all(d[..., 2] == 30.0)

    d2 = las.test_dim[np.array([0, 1, 10, 12])]
    assert np.all(d == d2)


def test_sub_field_view_with_self(simple_las_path):
    las = laspy.read(simple_las_path)

    rn = np.array(las.return_number)
    order = np.argsort(las.return_number)[::-1]

    las.return_number[:] = las.return_number[order]

    assert np.all(las.return_number == rn[order])


def test_can_use_array_func_with_list(simple_las_path):
    las = laspy.read(simple_las_path)

    np.concatenate([las.return_number, las.classification])
    np.concatenate([las.x, las.y])


def test_sub_field_as_array():
    array = np.zeros(10, np.uint8)
    field = SubFieldView(array, 0b0000_0010)

    cpy = np.array(field)

    cpy[:] = 1
    assert np.all(cpy == 1)
    assert np.all(field != 1)

    cpy[:] = 17

    with pytest.raises(OverflowError):
        field[:] = cpy[:]

    # Here we just test that this can run,
    # We had an error about with our __array__ signature
    cpy[:] = field[:]
    assert np.all(cpy == field)


def test_scaled_array_view():
    array = np.zeros(10, np.int32)
    x = ScaledArrayView(array, 0.01, 10)

    assert np.max(x) == 10.0
    assert np.min(x) == 10.0
    assert np.min(x) == x.min()
    assert np.max(x) == x.max()

    assert np.all(x > 0.0)
    assert np.all(x < 18.0)
    assert np.all(x == 10.0)
    assert np.all(x != 17.0)

    assert np.mean(x) == 10.0

    x[:] = 155.0
    x[9] = 42.0
    assert np.all(x[2:5] == 155.0)
    assert x[9] == 42.0

    with pytest.raises(OverflowError):
        x[8] = np.finfo(np.float64).max


def test_array_views_on_empty_things():
    """
    Test that __setitem__ of the Array views do not fail
    when the value is an empty array / sequence,
    to match the behaviour of numpy array
    """
    array = np.zeros(0, np.int32)
    x = ScaledArrayView(array, 0.01, 10)
    # This shall not fail
    x[:] = np.zeros(0)

    array = np.zeros(0, np.uint8)
    field = SubFieldView(array, 0b0000_0010)
    # This shall not fail
    field[:] = np.zeros(0)


def test_scaled_point_record_set_x_y_z():
    record = laspy.ScaleAwarePointRecord.zeros(
        5, point_format=laspy.PointFormat(3), scales=[1.0] * 3, offsets=[0.0] * 3
    )

    assert np.all(record.x == 0.0)
    assert np.all(record.y == 0.0)
    assert np.all(record.z == 0.0)

    record.x = 17.0
    record.y = 17.12
    record.z = np.array([293090.812739] * len(record))

    assert np.all(record.x == 17.0)
    assert np.all(record.y == 17.12)
    assert np.all(record.z == 293090.812739)



import io
from pathlib import Path

import laspy

simple_1_1 = Path(__file__).parent / "data" / "simple1_1.las"


def test_read_las_1_1():
    las = laspy.read(simple_1_1)


def test_create_las_1_1():
    las = laspy.create(point_format=1, file_version="1.1")


def write_las_1_1():
    las = laspy.create(point_format=1, file_version="1.1")
    with io.BytesIO() as out:
        las.write(out)
        out.seek(0)
        las = las.read(out)



import io
import os
from pathlib import Path
from typing import List, Tuple

import pytest

import laspy
from laspy import LasData, LasHeader, ScaleAwarePointRecord
from tests.conftest import NonSeekableStream
from tests.test_common import simple_laz


def test_append(file_path):
    """
    Test appending
    """
    if file_path.suffix == ".laz" and not laspy.LazBackend.Lazrs.is_available():
        pytest.skip("Only Lazrs backed supports appending")
    append_self_and_check(file_path)


def test_raises_for_laszip_backend():
    with pytest.raises(laspy.LaspyException):
        with laspy.open(simple_laz, mode="a", laz_backend=laspy.LazBackend.Laszip):
            ...


def test_append_las_with_evlrs():
    las = append_self_and_check(os.path.dirname(__file__) + "/data/1_4_w_evlr.las")

    expected_evlr = laspy.VLR(
        user_id="pylastest", record_id=42, description="just a test evlr"
    )
    expected_evlr.record_data = b"Test 1 2 ... 1 2"

    assert len(las.evlrs) == 1
    evlr = las.evlrs[0]
    assert evlr.description == expected_evlr.description
    assert evlr.record_id == expected_evlr.record_id
    assert evlr.user_id == expected_evlr.user_id
    assert evlr.record_data == expected_evlr.record_data


@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(), reason="Lazrs is not installed"
)
def test_append_laz_with_evlrs():
    las = append_self_and_check(os.path.dirname(__file__) + "/data/1_4_w_evlr.laz")

    expected_evlr = laspy.VLR(
        user_id="pylastest", record_id=42, description="just a test evlr"
    )
    expected_evlr.record_data = b"Test 1 2 ... 1 2"

    assert len(las.evlrs) == 1
    evlr = las.evlrs[0]
    assert evlr.description == expected_evlr.description
    assert evlr.record_id == expected_evlr.record_id
    assert evlr.user_id == expected_evlr.user_id
    assert evlr.record_data == expected_evlr.record_data


def append_self_and_check(las_path_fixture):
    with open(las_path_fixture, mode="rb") as f:
        file = io.BytesIO(f.read())
    las = laspy.read(las_path_fixture)
    print(las.vlrs)
    with laspy.open(file, mode="a", closefd=False) as laz_file:
        laz_file.append_points(las.points)
    file.seek(0, io.SEEK_SET)
    rlas = laspy.read(file)
    assert rlas.header.point_count == 2 * las.header.point_count
    assert rlas.points[: rlas.header.point_count // 2] == las.points
    assert rlas.points[rlas.header.point_count // 2 :] == las.points

    return rlas


def test_trying_to_append_in_non_seekable_raises():
    with pytest.raises(TypeError):
        with open(simple_laz, mode="rb") as f:
            stream = NonSeekableStream(f)
            with laspy.open(stream, mode="a") as lasf:
                pass


@pytest.mark.parametrize(
    "points_read_counts",
    [
        [-1],
        [1, -1],
        [49_999, -1],
        [50_000, -1],
        [50_001, -1],
    ],
)
@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(), reason="Lazrs is not installed"
)
def test_write_then_append_produces_valid_cloud(points_read_counts: List[int]) -> None:
    """
    In this test we read a LAZ file into chunks dictated by `points_read_counts`,
    then we reconstruct it by first writting the first chunk into a new file using write mode
    then we append the remaining chunks to that newly created file.

    At the end the input file and the one we patchworked must be the same
    """

    def iter_points(
        points_list: List[ScaleAwarePointRecord],
    ) -> Tuple[ScaleAwarePointRecord, slice]:
        """
        Returns a tuple that associates each point record in the input list
        with the slice to be used to recover the same chunk from
        the input file.
        """
        offset: int = 0

        for points in points_list:
            yield points, slice(offset, offset + len(points))
            offset += len(points)

    def check_write_append_read(
        points_list: List[ScaleAwarePointRecord], header: LasHeader
    ) -> None:
        """
        Implements the logic of first using write mode for the first
        chunk, then appending the remaining chunks.

        Checks result at the end
        """
        with io.BytesIO() as tmp_output:
            with laspy.open(
                tmp_output, "w", header=header, closefd=False, do_compress=True
            ) as las_writer:
                las_writer.write_points(points_list[0])

            for points in points_list[1:]:
                tmp_output.seek(0)
                if points:
                    with laspy.open(tmp_output, "a", closefd=False) as las_appender:
                        las_appender.append_points(points)

            tmp_output.seek(0)
            final_cloud: LasData = laspy.read(tmp_output)

            for points, selector in iter_points(points_list):
                assert final_cloud.points[selector] == points

    input_cloud_path = Path(__file__).parent / "data" / "autzen_trim.laz"

    with laspy.open(input_cloud_path) as las_reader:
        points_list: List[ScaleAwarePointRecord] = [
            las_reader.read_points(points_read_count)
            for points_read_count in points_read_counts
        ]

    check_write_append_read(points_list, las_reader.header)



import pytest

import laspy
from laspy import VLR

from .conftest import (
    LAS_1_4_WITH_EVLRS_FILE_PATH,
    LAZ_1_4_WITH_EVLRS_FILE_PATH,
    NonSeekableStream,
    las_file_path,
    laz_file_path,
)
from .test_chunk_read_write import check_chunked_reading_gives_expected_points

EXPECTED_EVLRS = [
    VLR(
        user_id="pylastest",
        record_id=42,
        description="just a test evlr",
        record_data=b"Test 1 2 ... 1 2",
    )
]


@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(),
    reason="None seekable laz is only supported by lazrs",
)
def test_laz_reading_non_seekable_stream(laz_file_path):
    with open(laz_file_path, mode="rb") as f:
        stream = NonSeekableStream(f)
        laspy.read(stream, closefd=False)


def test_las_reading_non_seekable_stream(las_file_path):
    with open(las_file_path, mode="rb") as f:
        stream = NonSeekableStream(f)
        laspy.read(stream, closefd=False)


@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(),
    reason="Non seekable laz is only supported by lazrs",
)
def test_laz_selective_decompression_reading_non_seekable_stream():
    with open(LAZ_1_4_WITH_EVLRS_FILE_PATH, mode="rb") as f:
        stream = NonSeekableStream(f)
        selection = laspy.DecompressionSelection.base()
        las = laspy.read(stream, closefd=False, decompression_selection=selection)
        assert las.evlrs == EXPECTED_EVLRS


@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(),
    reason="Non seekable laz is only supported by lazrs",
)
def test_laz_with_evlr_reading_non_seekable_stream():
    with open(LAZ_1_4_WITH_EVLRS_FILE_PATH, mode="rb") as f:
        stream = NonSeekableStream(f)
        las = laspy.read(stream, closefd=False)
        assert las.evlrs == EXPECTED_EVLRS


def test_las_with_evlr_reading_non_seekable_stream(las_file_path):
    with open(LAS_1_4_WITH_EVLRS_FILE_PATH, mode="rb") as f:
        stream = NonSeekableStream(f)
        las = laspy.read(stream, closefd=False)
        assert las.evlrs == EXPECTED_EVLRS


def test_non_seekable_chunked_las_reading(las_file_path):
    ground_truth = laspy.read(las_file_path)
    with open(las_file_path, mode="rb") as raw_file:
        las_reader = laspy.open(NonSeekableStream(raw_file), closefd=False)
        check_chunked_reading_gives_expected_points(
            ground_truth, las_reader, iter_size=130
        )


@pytest.mark.skipif(
    not laspy.LazBackend.Lazrs.is_available(),
    reason="Non seekable laz is only supported by lazrs",
)
def test_non_seekable_chunked_laz_reading(laz_file_path):
    ground_truth = laspy.read(laz_file_path)
    with open(laz_file_path, mode="rb") as raw_file:
        las_reader = laspy.open(NonSeekableStream(raw_file), closefd=False)
        check_chunked_reading_gives_expected_points(
            ground_truth, las_reader, iter_size=130
        )



import pytest


def skip_if_cli_deps_are_not_installed():
    try:
        from laspy.cli.core import app
    except ModuleNotFoundError:
        pytest.skip("skipping cli test (deps not installed)", allow_module_level=True)
